/*

  (C) 1997 Marco Falcioni (falcioni@ucla.edu)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
  USA.

  http://www.fsf.org/copyleft/gpl.html

  atoms methods.

  Comment: (1999) Links and atoms are inextricably interconnected

  Links are the entities that are used to access atoms which belong to
  a group for some reason (e.g., they belong to the same unique atom,
  or they are in the asymmetric unit, or they are in the same box in
  the grid...)
  Time comes when the same atom belongs to different groups, and is
  therefore referenced by more than one link. The issue is: when do I
  remove an atom? I must not remove an atom that is still referenced
  by an existing link. This would create inconsistencies and core
  dumps.  
  So the atom is only delete if the number of references to it is
  zero, and that is, when the last link to it has been removed.

  A minor design issue, due to the fact that we are not using C++, is
  that when new atoms are created they know nothing of the atom that
  has created them, and we cannot overload the creator.  I do not want
  to have two creators, so I have to do the copying by hand.
  

*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <strings.h>
#include <string.h>

#include "matrices.h"
#include "error.h"

#include "atoms.h"
#include "misc.h"
#include "cell.h"


long n_atoms = 0;

struct Atom * new_Atom(void) {
  struct Atom * A;
  A = (struct Atom *) Malloc(sizeof(struct Atom));

#ifdef DEBUG
  ++ n_atoms;
  printf("Atoms: %ld\n", n_atoms);
#endif

  A->ra = 0.0;
  A->rb = 0.0;
  A->rc = 0.0;
  A->rx = 0.0;
  A->ry = 0.0;
  A->rz = 0.0;

  /* the Debye-Waller factor */
  A->DW_fac = 1.0;
  /* the default occupancy */
  A->occupancy = 1.0;

  A->Radius = 1.0;

  A->merge_d = 0.0; 
  A->merge_n = 0;

  A->n_bonds = 0;
  A->mod = 1;
  A->consistent = 1; /* as funny as it may appear, so it is */
  A->SP_tag = 0;
  A->tag = 0;
  A->root = 0;
  A->refs = 0;
  strcpy(A->label, "");
  A->next = 0;
  return A;

}

void delete_Atom(struct Atom * A)
{
  if(A == 0) return;

#ifdef DEBUG
  -- n_atoms;
  printf("Atoms: %ld\n", n_atoms);
#endif

  if (A->refs != 0) {
    push_generic_error("delete_Atom","freeing Atom with non-zero refs");
  }

  Free(A);
  return;
}

void write_Atom(const struct Atom * A, FILE * stream)
{
  if(!A->consistent) 
    fprintf(stream," INCONSISTENT\n");
  fprintf(stream,"ABC %f %f %f\n",A->ra,A->rb,A->rc);
  fprintf(stream,"XYZ %f %f %f\n",A->rx,A->ry,A->rz);
  return;
}

void print_Atom(const struct Atom * A, FILE * stream)
{
  if(stream == 0) {
    push_generic_error("print_Atom","(nil) stream passed");
    exit_clean();
  }
  if(A == 0) {
    fprintf(stream,"Atom = (nil)\n");
    return;
  }
  fprintf(stream,"Atom at %p {\n",A);
  fprintf(stream,"ra = %f rb = %f rc = %f\n", A->ra, A->rb, A->rc);
  fprintf(stream,"rx = %f ry = %f rz = %f\n", A->rx, A->ry, A->rz);
  if(A->consistent != 1)  
    fprintf(stream, " INCONSISTENT");
  if(A->mod != 1)  
    fprintf(stream, " UNMODDED");
  if(A->merge_d > 0.0)  
    fprintf(stream, " MERGED by %f", A->merge_d);
  fprintf(stream, " label = %s",A->label);
  fprintf(stream,
	  "\nn_bonds = %d merge_n = %d refs = %d tag = %d\n", 
	  A->n_bonds, A->merge_n, A->refs, A->tag);
  fprintf(stream,"root = %d merge_d = %f merge_n = %d\n",
	  A->root, A->merge_d, A->merge_n);
  return;
}

void move_Atom(struct Atom * A, double * pos)
{
#ifndef NOCHECKS
  if(A == 0 || pos == 0) {
    push_generic_error("move_Atom","(nil) args passed");
    exit_clean();
  }
#endif
  A->mod = 0;
  A->consistent = 0;
  A->ra = pos[0];
  A->rb = pos[1];
  A->rc = pos[2];

  return;
}

void shift_Atom(struct Atom * A, double * pos)
{
#ifndef NOCHECKS
  if(A == 0 || pos == 0) {
    push_generic_error("shift_Atom","(nil) args passed");
    exit_clean();
  }
#endif
  A->mod = 0;
  A->consistent = 0;
  A->ra += pos[0];
  A->rb += pos[1];
  A->rc += pos[2];

  return;
}

void mod_Atom(struct Atom * A)
{
#ifndef NOCHECKS
  if(A == 0) {
    push_generic_error("mod_Atom","(nil) args passed");
    exit_clean();
  }
#endif
  if(A->mod == 1) return;
  A->mod = 1;
  A->consistent = 0;
  A->ra = fmod(fmod(A->ra, 1.0) + 1.0, 1.0);
  A->rb = fmod(fmod(A->rb, 1.0) + 1.0, 1.0);
  A->rc = fmod(fmod(A->rc, 1.0) + 1.0, 1.0);

  return;
}

void xyz_Atom(struct Atom * A, const struct Cell * C)
{
  double a[3],b[3];
#ifndef NOCHECKS
  if(C==0 || A==0) {
    push_generic_error("xyz_Atom","(nil) args passed");
    exit_clean();
  }
#endif
  if(A->consistent == 1) {
    return; /* nothing to do */
  }

  A->consistent = 1;
  a[0] = A->ra;
  a[1] = A->rb;
  a[2] = A->rc;
  matvec_multiply_d(b,C->abc_to_xyz,a,3);
  A->rx = b[0];
  A->ry = b[1];
  A->rz = b[2];

  return;
}


/* eventually we'll remove this */
void set_label_Atom(struct Atom * A, const char * string) 
{
#ifndef NOCHECKS
  if(A == 0 || string == 0) {
    push_generic_error("set_label_Atom","(nil) args passed");
    exit_clean();
  }
#endif
  strcpy(A->label, string);
  return;
}

/* copy the position independent quantities */

void clone_Atom(struct Atom * A, const struct Atom * B)
{
#ifndef NOCHECKS
  if(A == 0 || B == 0) {
    push_generic_error("clone_Atom","(nil) args passed");
    exit_clean();
  }
#endif
  if(A == B) return;

  A->SP_tag = B->SP_tag;
  A->n_bonds = B->n_bonds;
  A->root = B->root;
  A->DW_fac = B->DW_fac;
  A->occupancy = B->occupancy;
  A->Radius = B->Radius;
  set_label_Atom(A, B->label);

  return;

}
