/*

  (C) 1998 Marco Falcioni UCLA

  Build a PXD pattern starting from a given configuration, simmetry and cell
  it will compute multiplicities and clump near peaks together.

  usage:
       make_PXD < FAU.inp > fau.pxd
  
  it requires the generic.pxd file to be part of the input, or in the
  global directory for data files

  it outputs on stdout a file which can be used as input for zefsa.

  there are adjustable parameters: the angle cut (the max angle it should output)
  a lowest angle (low angle peaks are usually covered by the background)

  the composite range: expressed in degrees and referred to 2 theta, 
  it determines how far two peaks are to be still considered one.
  
  the output should be scanned for sanity!

*/


#include <stdio.h>
#include <stdlib.h>

/* thanks to Jerry */
#include "error.h"
#include "matrices.h"
#include "resources.h"
#include "stringlib.h"

#include "functions.h"
#include "grid.h"
#include "positions.h"
#include "cell.h"
#include "symmetry.h"
#include "rad_pattern.h"
#include "params.h"
#include "SP_entry.h"
#include "misc.h"
#include "link.h"
#include "oogl.h"
#include "chain.h"
#include "atoms.h"
#include "verbose.h"

int main (int argc, char ** argv)
{
  struct Cell * C = 0;
  struct Positions * T = 0;
  struct Sym_group * G = 0;
  struct Params * P = 0;
  struct Rad_pattern * RP = 0;
  struct string_list * options;
  int composites = 0, polar = 0, multi = 1;
  double a_max= 35.0, a_min= 5.0, sep = 0.02, i_min = 0.1;
  
  connect_error_stream(stderr, argv[0]);
 
  /* 
     get command line option: recognized options:
 
     v     verbose
     max   maximum angle
     min   minimum angle
     sep   separation for composite peaks
     comp  make composite peaks

  */

  options = append_string_to_list(0, "comp");
  options = append_string_to_list(options, "max");
  options = append_string_to_list(options, "min");
  options = append_string_to_list(options, "sep");
  options = append_string_to_list(options, "int");
  options = append_string_to_list(options, "polar");
  options = append_string_to_list(options, "nom");
  options = append_string_to_list(options, "v");

  args_to_resources(argc, argv, options);
    
  if(extract_global_resource("v")) connect_verbose_stream(stderr, argv[0]);
  if(extract_global_resource("polar")) polar = 1;
  if(extract_global_resource("comp")) composites = 1;
  if(extract_global_resource("nom")) multi = 0;
  if(composites && extract_global_resource("sep")) sep = double_from_resource("sep", 0.2);

  if(extract_global_resource("max")) a_max = double_from_resource("max", 35.0);
  if(extract_global_resource("int")) i_min = double_from_resource("int", 0.1);
  if(extract_global_resource("min")) a_min = double_from_resource("min", 5.0);

  P = read_Params(stdin);
  
  T = init_Positions(stdin);
  
  C = read_Cell(stdin);

  G = read_Sym_group(T->n_symmetry, stdin);

  if(C->csym == is_csym_Sym_group (G)) {
    /* no problem */
  } else {
    push_generic_error("read_data", "Centrosymmetric group/cell mismatch: check input");
    exit_clean();
  }
  read_Positions(T, stdin);
  
  P->SP = read_SP_file(T, P->Scat_Fact_file);

  if(P->Rad_file != 0) {
    FILE * a;
    RP = new_Rad_pattern();
    a = fopen(P->Rad_file, "r");
    if(a == 0) { /* try to read from the data path */
      char file[200];
      strcpy(file, P->data_path); 
      strcat(file, P->Rad_file);
      a = safe_open(file, "r"); /* if this does not work we bail out */
    }
    read_Rad_pattern(RP, a);
    fclose(a);
    get_correction(RP, T, C, P->SP);
  }

  test_Sym_group(G);
  if(polar) find_polar_Sym_group(G);

  symmetrize_Positions(T,G);
  mod_Positions(T);

  merge_positions(T,P,C);
  mod_Positions(T);

  angle_Rad_pattern(RP, C);

  get_intensity(RP,T,C, G);

  /* 
     we only want spots with intensities larger that the 
     third argument and with a 2 theta angle less than the
     fourth argument 
  */
  
  if (multi) find_multiplicity(RP, C, a_min , a_max, i_min);
  else copy_Rad_pattern(RP);

  sort_Rad_pattern(RP);
  if(composites)  find_composites(RP, sep); 
  norm_Rad_pattern(RP, ALL);
  write_Rad_pattern(RP,stdout); 

  delete_Rad_pattern(RP);
  delete_Positions(T);
  delete_Params(P);
  free_string_list(options);
  return 0;
  
}

