/*

  (C) Marco Falcioni UCLA 1998

  this routine takes a raw RP with a list of 
  computed intensies and tries to find the spots which are 
  related by simmetry in an empirical fashion.  It is unelegant but 
  I'd rather do this and not input 230 space groups.

*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <malloc.h>

#include "error.h"
#include "constants.h"
#include "memory.h"

#include "cell.h"
#include "misc.h"
#include "rad_pattern.h"
#include "spot.h"
#include "intensity.h"

#define MULT
#define NOCANC

void find_multiplicity(struct Rad_pattern * RP, const struct Cell * C,
		       double min_angle, double max_angle, double min_int) 
{
  static double tiny = 1.0E-5;
  struct Spot ** New_Spots;
  struct Spot ** tmp;
  unsigned new_length = 0;
  unsigned i;

  if(RP == 0 || min_angle < 0.0 || max_angle < min_angle) {
    push_generic_error("find_multiplicity", "(nil) or weird args passed, you dope");
    exit_clean();
  }

/*****************************************************************************\
 *                                                                           *
 *     Scale the intensities so that the highest is 1000.0                   *
 *     this is to make the intensity cut equal                               *
 *     Also, compute the angles                                              *
 *                                                                           *
\*****************************************************************************/

  norm_Rad_pattern(RP, CALC);

  angle_Rad_pattern(RP, C);

/*****************************************************************************\
 *                                                                           *
 *   Find the multiplicities for the spots that fit the criteria             *
 *   of angle and intensity                                                  *
 *                                                                           *
\*****************************************************************************/

  New_Spots = (struct Spot **) Malloc(RP->length * sizeof(struct Spot *));

  for(i = 0; i < RP->length; i ++) { /* main loop over spots */
    struct Spot * aa = RP->I[i];



    /* enforce cut */
    if(aa->twotheta < min_angle || aa->twotheta > max_angle) continue;

#ifdef DEBUG
      printf("looking at %f %d %d %d %f\n", RP->I[i]->twotheta, (int) RP->I[i]->h_, (int) RP->I[i]->k_,  
	     (int) RP->I[i]->l_, RP->I[i]->xray->calc);

#endif

    /* remove cancellations */

#ifdef NOCANC
    switch (RP->type) {
    case XRAY:
      if(RP->I[i]->xray->calc < min_int ) continue;
      break;
    case NEUTRON:
      if(RP->I[i]->neutron->calc < min_int ) continue;
      break;
    case BOTH:
      if(RP->I[i]->neutron->calc < min_int && RP->I[i]->xray->calc < min_int ) continue;
      break;
    case NONE:
      /* fall through */
    default:
      push_generic_error("find_multiplicity","weirdness in enum switch");
      exit_clean();
    }
#endif

    if(new_length == 0) {
      New_Spots[0] = new_Spot(RP->type);
      copy_Spot(RP->I[i], New_Spots[0]);
      new_length = 1;
    } else {
      unsigned found = 0;
      unsigned j;



      for(j = 0; j < new_length; j ++) {

	if(fabs(RP->I[i]->twotheta - New_Spots[j]->twotheta) < tiny)
	switch (RP->type) {
	case XRAY:
	  if(fabs(RP->I[i]->xray->calc - New_Spots[j]->xray->calc) / 
	     (RP->I[i]->xray->calc + New_Spots[j]->xray->calc) < tiny) {
	    New_Spots[j]->multiplicity++;
	    found ++;
	  }
	  break;
	case NEUTRON: 
	  if(fabs(RP->I[i]->neutron->calc - New_Spots[j]->neutron->calc)/
	     fabs(RP->I[i]->neutron->calc + New_Spots[j]->neutron->calc) < tiny) {
	    New_Spots[j]->multiplicity++;
	    found ++;
	  }
	  break;
	case BOTH:
	  push_generic_error("find_multiplicity",
			     "cannot deal with both radiations FEATURE NOT IMPLEMENTED");
	  exit_clean();
	  break;
	case NONE:
	  /* fall through */
	default:
	  push_generic_error("find_multiplicity","weirdness in enum switch");
	  exit_clean();
	}
	if(found) goto isfound;
      }
    isfound:
      if(!found) {
	New_Spots[new_length] = new_Spot(RP->type);
	copy_Spot( RP->I[i], New_Spots[new_length]);
	new_length++;
#ifdef DEBUG
	printf("adding a spot %f %f %f\n", RP->I[i]->h_, RP->I[i]->k_, RP->I[i]->l_);
#endif
      } else {
#ifdef DEBUG
	printf("found in list\n");
#endif
      }
    }
    
  }
  
  /* put new spots in place of the old ones and update the length */
  
  tmp = RP->I;
  RP->I = New_Spots;
  RP->length = new_length;
  
  /***************************************************************************\
   *                                                                         *
   *  include multiplicity and scale spots once more!                        *
   *                                                                         *
  \***************************************************************************/
  
#ifdef MULT  
  for(i = 0; i < RP->length; i ++) {
    switch (RP->type) {
    case XRAY:
      RP->I[i]->xray->calc *= RP->I[i]->multiplicity;
      break;
    case NEUTRON:
      RP->I[i]->neutron->calc *= RP->I[i]->multiplicity;
      break;
    case BOTH:
      RP->I[i]->xray->calc *= RP->I[i]->multiplicity;
      RP->I[i]->neutron->calc *= RP->I[i]->multiplicity;
      break;
    case NONE:
      /* fall through */
    default:
      push_generic_error("find_multiplicity",
			 "weirdness in second enum switch");
      exit_clean();
    }
  }
#endif

  norm_Rad_pattern(RP, CALC);
  
  /* transfer computed values to value for printing */

  for(i = 0; i < new_length; i ++) {
    switch (RP->type) {
    case XRAY:
      RP->I[i]->xray->value = RP->I[i]->xray->calc; 
      break;
    case NEUTRON:
      RP->I[i]->neutron->value = RP->I[i]->neutron->calc; 
      break;
    case BOTH:
      RP->I[i]->xray->value = RP->I[i]->xray->calc; 
      RP->I[i]->neutron->value = RP->I[i]->neutron->calc; 
      break;
    case NONE:
      /* fall through */
    default:
      push_generic_error("find_multiplicity",
			 "weirdness in second enum switch");
      exit_clean();
    }
  }

  

  for(i = 0; i < RP->length; i++) {
    delete_Spot(tmp[i]);
  }

  free(tmp);

  return;

}
