/*

  (C) 1997 Marco Falcioni (falcioni@ucla.edu)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
  USA.

  http://www.fsf.org/copyleft/gpl.html


*/
/*  
    (C) 1997 Marco Falcioni

    routine to merge atoms which are closer than a given 
    distance.  The merged atom will have a position averaged over the 
    positions of all the other ones.  Only atoms related by a symmetry 
    operation can be merged.


    2/1999

    removed potential memory leak (see ***3)

*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "matrices.h"

#include "error.h"
#include "positions.h"
#include "cell.h"
#include "chain.h"
#include "link.h"
#include "atoms.h"
#include "params.h"
#include "misc.h"

void merge_positions(struct Positions * T,
		     const struct Params * P,
		     const struct Cell * C)
{
  unsigned long i,k;
  unsigned T_density = 0;
  /* this seems a good idea */
  /* create chains for each atom and keep a handle to them */
  struct Chain ** hood;

#ifndef NOCHECKS
  if(T == 0 || C == 0) {
    push_generic_error("merge_positions","(nil) args passed");
    exit_clean();
  }
#endif

  if(T->test == 0 || T->mod == 0) {
    push_generic_error("merge_positions","positions not symmetric or moded");
    exit_clean();
  }

  /* ***3 the following two statements were in opposite order */

  if(T->merged == 1) return;

  hood = (struct Chain **) Malloc(T->n_symmetry * sizeof(struct Chain *));

  clean_Chain(T->distinct);

  /* it is convenient to think in terms of symmetry related atoms */
  for(i = 0; i < T->n_unique; i++) {
    unsigned long j = 0;
    struct Link * pp;
#ifdef DEBUG
    printf("Unique atom %ld\n",i);
#endif

    /* 
       we need a chain (and link) to each atom in the list;
       this allows us to concatenate atoms which are close enough
       to be merged 
    */

    for(pp = T->sym_related[i]->head; pp != 0; pp = pp->next, j++) {
      struct Link * qq;

#ifdef DEBUG  
      printf("Sym Related atoms: Init %ld\n", j);
#endif

      pp->atom->merge_d = -1.0;
      pp->atom->merge_n = 0;
      hood[j] = new_Chain();
      qq = new_Link(pp->atom);
      add_Link(qq,hood[j]);
    }
  
    /* go through the atoms symmetry wise */
  
    j = 0;
    for(pp = T->sym_related[i]->head; pp != 0; pp = pp->next, j++) {
      struct Link * bb;
      long k = 0;

#ifdef DEBUG
      printf("\nSym Related atoms: Loop I %ld: ", j);
#endif

      for(bb = T->sym_related[i]->head; bb != 0; bb = bb->next, k++) {
	double d2;

#ifdef DEBUG
	printf("%ld ", k);
#endif

	if(pp == bb || hood[j] == 0) continue;
	d2 = min_dist2_xyz(C, pp->atom, bb->atom);
	if(d2 < P->merge2) {
#ifdef DEBUG
	  printf("\ndistance %f < %f? real %f\n",d2, 
		 P->merge2, dist2_xyz(C, pp->atom, bb->atom));
#endif

	  hood[j] = cat_Chain(hood[j],hood[k]);
	  hood[k] = 0;
	}
      }
    }
  
  /* 
     now we actually merge the atoms: the new position is the average 
     since averaging is a linear operation we do not need to do it in 
     real space
   */

    for(j = 0; j < T->n_symmetry; j++) {
      double ave[3], a[3];
      unsigned long length;
      double invl;
      struct Link * aa, * bb;

#ifdef DEBUG
	  printf("\nMerging %ld: ", j);
#endif

      if(hood[j] == 0) continue;
      
      if(hood[j]->length >= 1) {
#ifdef DEBUG
	  printf("Is Distinct %ld len %ld\n", j, hood[j]->length);
#endif
	aa = new_Link(hood[j]->head->atom);
	add_Link(aa, T->distinct); /* that's all! */
	T_density++;
      }
      
      if(hood[j]->length <= 1) { 
	clean_Chain(hood[j]);
	delete_Chain(hood[j]);
	hood[j] = 0;
	continue;
      }
      
      length = hood[j]->length;
      invl = 1.0/(double) length;
      
      /* here we compute the average position of mergee atoms */
      
      aa = hood[j]->head;
      
      a[0] = ave[0] = aa->atom->ra;
      a[1] = ave[1] = aa->atom->rb;
      a[2] = ave[2] = aa->atom->rc;
      
      for(bb = aa->next; bb != 0; bb = bb->next) {
	double b[3];
 
	b[0] = bb->atom->ra;
	b[1] = bb->atom->rb;
	b[2] = bb->atom->rc;
	
	for(k = 0; k < 3; k++) {
	  double C = a[k] - b[k];
	  if( C > 0.5 ) b[k] += 1.0;
	  if( C < -0.5 ) b[k] -= 1.0;
	  ave[k] += b[k];
	  a[k] = b[k];
	}
      }
      
      /* we mod the average position in the unit cell */
      for(k = 0; k < 3; k++) {
	ave[k] *= invl;
	ave[k] = fmod(fmod(ave[k],1.0)+1.0,1.0);
      }
      
      /* now actually move all the atoms in the chain */
      for(aa = hood[j]->head; aa != 0; aa = aa->next) {
	double pos[3], abc[3];
	/* here we tag the atom with the distance by which it is moved */

	a[0] = aa->atom->ra;
	a[1] = aa->atom->rb;
	a[2] = aa->atom->rc;	

	for(k = 0; k < 3; k++) {
	  double v = ave[k] - a[k];
	  if( v > 0.5 ) v -= 1.0;
	  if( v < -0.5 ) v += 1.0;
	  abc[k] = v;
	}

#ifdef DEBUG
	printf("ABC %f %f %f\n", abc[0], abc[1], abc[2]);
	write_Atom(aa->atom, stdout);
#endif

	matvec_multiply_d(pos, C->abc_to_xyz, abc, 3);

	aa->atom->merge_d = dot(pos,pos); 
	aa->atom->merge_n = (unsigned) hood[j]->length;

	
	move_Atom(aa->atom, ave);
#ifdef DEBUG
	printf("merge_d = %f\n", aa->atom->merge_d);
	write_Atom(aa->atom, stdout);
#endif
	aa->atom->mod = 1;
      }
      clean_Chain(hood[j]);
      delete_Chain(hood[j]);
      hood[j] = 0;
    }
  }

  Free(hood);

  T->density = T_density;
  T->merged = 1;
  T->mod = 1;
  T->n_actual = T->distinct->length;
  return;
}

