/*

  Marco Falcioni (C) 1997

  miscellaneous routines

*/
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <math.h>


#include "misc.h"

extern int errno;


/* from lingjm-1.3 */
#include "memory.h"
#include "error.h"




/*---------------------------------------------------------------------------*
 *        Strips whitespace and #type comments                               *
 *---------------------------------------------------------------------------*/
char * clean_line(char * q)
{
  char * p = (char *) q;
  char * ret;

  /* safety check */
  if(p == 0) return 0;
  /* eat whitespace */
  while(isspace((int) p[0]) && p[0] != '\0') p++;


  /* skip comments, empty lines */
  if(p[0] == '#' || p[0] == '\0') {
    return 0;
  } 
  ret = p;

  /* remove trailing comments */
  for(; (*p) != '\0'; p++) if( *p == '#') *p = '\0'; 
  
  return ret;
}

/*---------------------------------------------------------------------------*
 *            count the number of columns in a given input line  (see below) *
 *---------------------------------------------------------------------------*/
int count_cols (const char * line)
{
  int state = 1;
  int cols = 1;
  
  if(line == 0) return 0; /* trap */
  
  /* eat leading whitespace */
  while(isspace((int) *line) && *line != '\0') line++;
  

  /* the line was just whitespace or comments*/
  if(*line == '\n' || *line == '\0' || *line == '#') return 0;

  state = 0; /* there is at least one non white character */
  
  while(*line != '\0' && *line != '#'){ /* both terminate line */
    int c = isspace((int) *line); /* c = 0 -> non white char */
    if(c != 0) { /* a word has ended */
      state = 1;
    } else if (c == 0 && state == 1) { /* a word begun: new column */
      cols++;
      state = 0;
    }
    line++;
  }
  return cols;
}
/*---------------------------------------------------------------------------*
 *           read n (long) integers from a given line (see above)            *
 *---------------------------------------------------------------------------*/
long * read_longs(unsigned long n, const char * line)
{
  unsigned long i;
  char * ptr;
  char * tmp = (char *) line;
  long * result;

    if(line == 0) return 0; /* trap gremlins */

  result = new_vector_l(n);

  for(i=0; i<n; i++) {
    result[i] = strtol(tmp, &ptr, 10); /* base 10 */
    tmp = ptr;
  }
  return result;
}
/*---------------------------------------------------------------------------*
 *        get a line from a stream and be wordy about errors                 *
 *---------------------------------------------------------------------------*/
char * get_line (char * buf, int bufsize, FILE * stream)
{
  char * res = 0;
  errno = 0;
  do {
    fgets(buf, bufsize, stream);
    if(feof(stream)) {
      push_generic_error("get_line",
			 "EOF seen in stream... hope you catch that");
      return 0;
    } else if (ferror(stream)) {
      perror("get_line");
      exit(errno);
    }
    res = clean_line(buf);
  } while (res == 0);

  return res;

}


/*---------------------------------------------------------------------------*
 *          open a file and be wordy if errors occur                         *
 *---------------------------------------------------------------------------*/

FILE * safe_open(char * file, char * mode)
{
  FILE * stream;
  
  if(file == (char *)0 || mode == (char *) 0){
    push_generic_error("safe_open","(nil) args passed");
    exit_clean();
  }
  stream = fopen(file,mode);
  if(stream == (FILE *)0){
    char buf[1024];
    sprintf(buf, "%s: %s", file, strerror(errno));
    push_generic_error("safe_open",buf);
    exit_clean();
  }
  return stream;
}

/*---------------------------------------------------------------------------*
 *          burp out information about the current execution environment     *
 *---------------------------------------------------------------------------*/
void run_ID (FILE * stream)
{
  char * user;
  char * pwd;
  char * host;
  char * hosttype;
  char * date;
  time_t t;


  if(stream == 0) {
    push_generic_error("run_ID","(nil) stream passed");
    exit_clean();
  }

  t = time(NULL);
  date = asctime(localtime(&t));  

  user = getenv("USER");
  pwd = getenv("PWD");
  host = getenv("HOST");
  hosttype = getenv("HOSTTYPE");

  fprintf(stream, "# (C) 1997 Marco Falcioni & Michael W. Deem UCLA\n"); 
  fprintf(stream, "# executed on %s", date); 
  fprintf(stream, "# by %s on %s (%s)\n", user, host, hosttype);
  fprintf(stream, "# the current directory is %s\n", pwd);

  
  if(ferror(stream)) {
    push_generic_error("run_ID","error while writing to stream");
    exit_clean();
  }

  return;

}

/*---------------------------------------------------------------------------*
 *           return a random gaussian deviate                                *
 *---------------------------------------------------------------------------*/
double gasdev(void)
{
  static int iset = 0;
  static double gset;
  double v1, v2, rsq, fac;

  if(iset == 0) {
    do {
      v1 = 2.0 * Random() - 1.0;
      v2 = 2.0 * Random() - 1.0;
      rsq = v1 * v1 + v2 * v2;
    } while (rsq >= 1.0 || rsq == 0);
    fac = sqrt(-2.0 * log(rsq) / rsq );
    gset = v1 * fac;
    iset = 1;
    return v2 * fac;
  } else {
    iset = 0;
    return gset;
  }
}
