/*

  (C) 1999 Marco Falcioni (falcioni@ucla.edu)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
  USA.

  http://www.fsf.org/copyleft/gpl.html


*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <strings.h>
#include <string.h>

#include "matrices.h"
#include "error.h"
#include "constants.h"

#include "atoms.h"
#include "vector_functions.h"
#include "molecule.h"
#include "misc.h"
#include "cell.h"
#include "trial.h"
#include "params.h"



extern const int LINELENGTH;


struct Molecule * new_Molecule(int N) 
{
  struct Molecule * M;
  int i;
  M = (struct Molecule *) Malloc(sizeof(struct Molecule));

  M->N = N; 

  M->com[0] = 0.0;
  M->com[1] = 0.0;
  M->com[2] = 0.0;

  M->A = (struct Atom **) malloc(N * sizeof(struct Atom *));
  for(i = 0; i < M->N; i++) {
    M->A[i] = new_Atom();
  }
  
  M->phi_M = 0.0;
  M->phi_m = 0.0;
  M->phi = 0.0;

  return M;

}

struct Molecule * read_Molecule(FILE * stream) 
{
  int i, c, N;
  double phiM, phim, DW, occ;
  struct Molecule * M;
  char *line, *p, word[1024];
  
  line = (char *) malloc(LINELENGTH);

  p = get_line(line, LINELENGTH , stream);
  c = sscanf(p, "%s%lf%lf%d", word, &(phiM), &(phim), &N);

  if(strncmp("MOLECULE", word, 8) != 0) {
    return(0);
  }
  
  if(N < 0) {
    push_generic_error("read_Molecule", "bad number of atoms");
    exit_clean();
  }

  M = new_Molecule(N);

  M->phi_M = phiM * constPi_/180.0;
  M->phi_m = phim * constPi_/180.0;
  M->phi = sqrt(M->phi_M * M->phi_m);

  for(i = 0; i < N; i++) {
    double x[3];
    p = get_line(line, LINELENGTH , stream);
    c = sscanf(p, "%s %lf %lf %lf %lf %lf", word, x, x+1, x+2, &DW, &occ);
    M->A[i]->rx = x[0];
    M->A[i]->ry = x[1];
    M->A[i]->rz = x[2];
    set_label_Atom(M->A[i], word);
    if(c >= 5) M->A[i]->DW_fac = DW;
    if(c >= 6) M->A[i]->occupancy = occ;

  }

  free(line);

  return(M);

}


void delete_Molecule(struct Molecule * M)
{
  int i;
  if(M == 0) return;

  for(i = 0; i < M->N; i++) {
    delete_Atom(M->A[i]);
  }
  free(M->A);
  Free(M);
  return;
}

void print_Molecule(const struct Molecule * M, FILE * stream)
{
  int i;
  if(stream == 0) {
    push_generic_error("print_Molecule","(nil) stream passed");
    exit_clean();
  }

  if(M == 0) {
    fprintf(stream,"Molecule = (nil)\n");
    return;
  }

  /*  fprintf(stream,"Molecule at %p {\n", M);
      fprintf(stream,"has %d atoms\n", M->N); */
  for(i = 0; i < M->N; i++) {
    fprintf(stream, "%s %f %f %f\n", M->A[i]->label, M->A[i]->rx,
	    M->A[i]->ry,  M->A[i]->rz); 
  }
  /* fprintf(stream,"Phi_M %f Phi_m %f\n", M->phi_M, M->phi_m); 
     fprintf(stream,"}\n"); */

  return;
}

void write_Molecule_atoms(const int Nmol,
                       struct Molecule ** M, FILE * stream)
{
  int i, j;

    for (i = 0; i < Nmol; i++) {
      for (j = 0; j < M[i]->N; j++) {
          fprintf(stream, "# %4s %f %f %f %7.4f %7.4f\n",
            M[i]->A[j]->label,
            M[i]->A[j]->rx,
            M[i]->A[j]->ry,
            M[i]->A[j]->rz,
            M[i]->A[j]->DW_fac,
            M[i]->A[j]->occupancy);

      }
    }
  return;
}


void com_Molecule(struct Molecule * M)
{
  int i;
#ifndef NOCHECKS
  if(M == 0) {
    push_generic_error("move_Molecule","(nil) args passed");
    exit_clean();
  }
#endif

  M->com[0] = M->A[0]->rx;
  M->com[1] = M->A[0]->ry;
  M->com[2] = M->A[0]->rz;

  for(i = 1; i < M->N; i++) {
    M->com[0] += M->A[i]->rx;
    M->com[1] += M->A[i]->ry;
    M->com[2] += M->A[i]->rz;
  }

  M->com[0] /= (double) M->N;
  M->com[1] /= (double) M->N;
  M->com[2] /= (double) M->N;

  return;
}

void shift_Molecule(struct Molecule * M, struct Trial * T)
{
  int i;

#ifndef NOCHECKS
  if(M == 0 || T == 0) {
    push_generic_error("shift_Molecule","(nil) args passed");
    exit_clean();
  }
#endif

  for(i = 0; i < M->N; i++) {
    M->A[i]->consistent = 0;
    M->A[i]->rx += T->dx;
    M->A[i]->ry += T->dy;
    M->A[i]->rz += T->dz;
  }

  return;
}

void rotate_Molecule(struct Molecule * M, double angle, double * dir)
{
  int i;
  double pos[3], newpos[3];

#ifndef NOCHECKS
  if(M == 0) {
    push_generic_error("rotate_Molecule","(nil) args passed");
    exit_clean();
  }
#endif

  com_Molecule(M); /* update center of mass */
  
  /* shift molecule at origin */

  for(i = 0; i < M->N; i++) {
    M->A[i]->rx -= M->com[0];
    M->A[i]->ry -= M->com[1];
    M->A[i]->rz -= M->com[2];
  }
  
  for(i = 0; i < M->N; i++) {
    pos[0] = M->A[i]->rx;
    pos[1] = M->A[i]->ry;
    pos[2] = M->A[i]->rz;
    
    rotate_vector(pos, dir, newpos, cos(angle), sin(angle)); 

    M->A[i]->rx = newpos[0];
    M->A[i]->ry = newpos[1];
    M->A[i]->rz = newpos[2];
  }

  /* shift molecule at back */

  for(i = 0; i < M->N; i++) {
    M->A[i]->rx += M->com[0];
    M->A[i]->ry += M->com[1];
    M->A[i]->rz += M->com[2];
  }

  return;
}

void abc_Molecule(struct Molecule * M, const struct Cell * C)
{
  int i;
  double a[3],b[3];
#ifndef NOCHECKS
  if(C == 0 || M == 0) {
    push_generic_error("abc_Molecule","(nil) args passed");
    exit_clean();
  }
#endif

  for(i = 0; i < M->N; i++) {
    a[0] = M->A[i]->rx;
    a[1] = M->A[i]->ry;
    a[2] = M->A[i]->rz;
    matvec_multiply_d(b, C->xyz_to_abc, a, 3);
    M->A[i]->ra = b[0];
    M->A[i]->rb = b[1];
    M->A[i]->rc = b[2];
    M->A[i]->consistent = 1;
  }

  return;
}


void mod_all_Molecules(const int Nmol,
                       struct Molecule ** M, const struct Cell * C)
{
  int i;

  for (i = 0; i < Nmol; i++)  mod_Molecule(M[i], C);
  return;
}

void mod_Molecule(struct Molecule * M, const struct Cell * C)
{
  int i;
  double a[3],b[3];
#ifndef NOCHECKS
  if(C == 0 || M == 0) {
    push_generic_error("abc_Molecule","(nil) args passed");
    exit_clean();
  }
#endif

  com_Molecule(M);

  a[0] = M->com[0];
  a[1] = M->com[1];
  a[2] = M->com[2];

  matvec_multiply_d(b, C->xyz_to_abc, a, 3);

/* MWD Change */
  b[0] = fmod(fmod(b[0], 1.0) + 1.0, 1.0);
  b[1] = fmod(fmod(b[1], 1.0) + 1.0, 1.0);
  b[2] = fmod(fmod(b[2], 1.0) + 1.0, 1.0);

  matvec_multiply_d(a, C->abc_to_xyz, b, 3);

  a[0] -= M->com[0];
  a[1] -= M->com[1];
  a[2] -= M->com[2];

  for(i = 0; i < M->N; i++) {
    M->A[i]->consistent = 0;
    M->A[i]->rx += a[0];
    M->A[i]->ry += a[1];
    M->A[i]->rz += a[2];
  }

  return;
}


void copy_Molecule(const struct Molecule * A, struct Molecule * B)
{
  int i;
#ifndef NOCHECKS
  if(A == 0 || B == 0) {
    push_generic_error("copy_Molecule","(nil) args passed");
    exit_clean();
  }
#endif

  for(i = 0; i < A->N; i++) {
    B->A[i]->rx = A->A[i]->rx;
    B->A[i]->ry = A->A[i]->ry;
    B->A[i]->rz = A->A[i]->rz;
    clone_Atom(B->A[i], A->A[i]);
  }

  B->com[0] = A->com[0];
  B->com[1] = A->com[1];
  B->com[2] = A->com[2];

  B->phi_M = A->phi_M; 
  B->phi_m = A->phi_m;
  B->phi = A->phi;

}

int set_phi(struct Molecule * M, struct Params * P, double fraction ) 
{
#ifndef NOCHECKS
  if(M == 0 || P == 0 || fraction < 0.0 ) {
    push_generic_error("set_phi","(nil) arg passed");
    exit_clean();
  }
#endif
  
  M->phi *= 1 + P->gamma * (fraction - P->mc_frac);

  if(M->phi < M->phi_m) {
    M->phi = M->phi_m;
    return 1;
  }
  /* this is a safety measure to prevent blowups */
  if(M->phi > M->phi_M) {
    M->phi = M->phi_M;
    return 1;
  }
  return 0;
}
