/*

  (C) 1998 Marco Falcioni UCLA


  Parallel tempering.


  Input: file "temperatures"
         format:
	 %d   # number n of temperatures
	 %lf %lf %lf [%lf] \
	 %lf %lf %lf [%lf] |
                           |
	 ... ... ...       >  n temperatures move amplitudes (radii) probs
	                   |                          and phi if
			   |			      molecules are there      
                           |
         %lf %lf %lf [%lf] |
	 %lf %lf %lf [%lf] /
	 %lf  # ratio of updates to temperature swap attempts    
	 

	 

  (C) 1997 Marco Falcioni (falcioni@ucla.edu)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
  USA.

  http://www.fsf.org/copyleft/gpl.html




*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "error.h"
#include "constants.h"

#include "atoms.h"
#include "chain.h"
#include "grid.h"
#include "conf.h"
#include "positions.h"
#include "link.h"
#include "misc.h"
#include "oogl.h"
#include "trial.h"
#include "params.h"
#include "functions.h"
#include "rad_pattern.h"
#include "verbose.h"
#include "macros.h"
#include "molecule.h"

#define ENERGY
#define OOGL

struct Cell;
struct Positions;
struct Rad_pattern;
struct Sym_group;

unsigned pick_system(double *, unsigned);

void parallel_tempering (struct Params * P, const struct Cell * C,
	     const struct Sym_group * G, struct Rad_pattern * RP,
	     struct Positions * T1)
{
  struct Conf ** Current;
  struct Grid * Gr;
  struct Positions ** TS = 0;
  /* 
     betas holds the tempering temperatures 
     radii holds the amplitude for move proposal
     fractions is quite useless
     rel_probs holds the relative probability of updating a given system
  */
  double * betas, * radii, * fractions, *rel_probs, p_norm, *phi;
  unsigned n_betas, i;
  FILE * Tmp_stream;
  char line[1024], message[1024], *p;
  double update_ratio;  /* the ratio at which we update vs. swap the structures */

#ifdef OOGL
  FILE ** oogl_streams;
#endif
#ifdef ENERGY
  FILE ** ene_streams;
  FILE ** mol_streams = 0;
#endif

  /* set up the parallel tempering scheme */

  Tmp_stream = safe_open("temperatures", "r");

  /* read in the first line: number of temperatures */
  p = get_line(line, sizeof(line), Tmp_stream);
  
  if(sscanf(p, "%d", &n_betas) > 1 || n_betas < 3) {
    push_generic_error("parallel_tempering", 
		       "the temperatures file is garbled! A");
    exit_clean();
  }
  push_verbose_message("parallel_tempering","reading temperatures file");  
  /* arrays */
 
  betas = (double *) Malloc(n_betas * sizeof(double));
  radii = (double *) Malloc(n_betas * sizeof(double));
  fractions = (double *) Malloc(n_betas * sizeof(double));
  rel_probs = (double *) Malloc(n_betas * sizeof(double));
  phi = (double *) Malloc(n_betas * sizeof(double));

#ifdef OOGL
  oogl_streams = (FILE **) Malloc(n_betas * sizeof(FILE *));
#endif
#ifdef ENERGY
  ene_streams = (FILE **) Malloc(n_betas * sizeof(FILE *));
  if(T1->Nmol) 
    mol_streams = (FILE **) Malloc(n_betas * sizeof(FILE *));
#endif

  TS = (struct Positions **) Malloc(n_betas * sizeof(struct Positions *));
  Current = (struct Conf **) Malloc(n_betas * sizeof(struct Conf *));

  symmetrize_Positions(T1, G);

  p_norm = 0.0; /* this is to normalize probabilities */

  for(i = 0; i < n_betas; i++) {
    int c;
    double ppp;
    char name[1024];
#ifdef OOGL
    sprintf(name, "hist_%d.oogl",i);
    oogl_streams[i] = safe_open(name, "w+");
#endif
#ifdef ENERGY
    sprintf(name, "hist_%d.ene",i);
    ene_streams[i] = safe_open(name, "w+");
    sprintf(name, "pos_%d.mol",i);
    if(T1->Nmol) mol_streams[i] = safe_open(name, "w+");
#endif

    fractions[i] = 0.0;
    p = get_line(line, sizeof(line), Tmp_stream);
    c = sscanf(line, "%lf%lf%lf%lf", 
	      &(betas[i]), &(radii[i]), &(rel_probs[i]), &ppp);
    if(c < 3 || (T1->Nmol > 0 && c < 4)) {
      push_generic_error("parallel_tempering",
			 "error reading temperature file B");
      exit_clean();
    }
    phi[i] = ppp * constPi_ / 180.0; /* phi[i] = 0 if not present */

    /* normalize probabilities */
    p_norm += rel_probs[i];

    TS[i] = new_Positions(T1->n_unique, T1->n_symmetry, T1->n_actual,
                          T1->Nmol, T1->M);
    clone_Positions(T1,TS[i]);
    /* we need T1 to get the correct atom labels, but then we randomize
       the positons to decorrelate the systems */
    rand_Positions(TS[i]);
    symmetrize_Positions(TS[i], G);
    mod_Positions(TS[i]);
    Current[i] = new_Conf(T1->n_unique);
  }

  /* normalize probabilities to one */
  for(i = 0; i < n_betas; i++) {
    rel_probs[i] /= p_norm;
  }

  /* read the ratio of updates to swaps */
  p = get_line(line, sizeof(line), Tmp_stream);
  if(sscanf(line, "%lf", &(update_ratio)) != 1 || p == 0 ) {
    push_generic_error("parallel_tempering",
		       "error reading temperature file C");
    exit_clean();
  }
  
  Gr = new_Grid(P, C, 0);
  
  /* MWD Change */
   /* assign phi values to all molecules */  
   { int j;
     for (j = 0; j < n_betas; j++) {
       for(i = 0; i < TS[j]->Nmol; i++) {
         TS[j]->M[i]->phi = phi[j];
       }
     }
   }

/* MWD Change */
  /* initialize energies */
  for(i = 0 ; i < n_betas; i++) {
    P->radius = radii[i];
    P->temperature = betas[i];
    symmetrize_Positions(TS[i],G);
    mod_Positions(TS[i]);
    merge_positions(TS[i],P,C);
    clean_Grid(Gr);
    fill_Grid(Gr,TS[i],C);
    Current[i]->tot_energy = energy (P, C, TS[i], Gr, Current[i]);
    Current[i]->RP_energy = 0.0;
    if(RP) { /* only if we have a RP */
      get_intensity(RP, TS[i], C, G);
      scale_Rad_pattern(RP);
      Current[i]->RP_energy = distance_Rad_pattern(RP);
      Current[i]->tot_energy += P->wt_radpat * Current[i]->RP_energy;
    }
  }
  for(i = 0 ; i < n_betas; i++) {
#ifdef ENERGY
      fprintf(ene_streams[i],"%d ", 0);
      fprint_Conf(Current[i], ene_streams[i]);
#endif
#ifdef OOGL
      OOGLCONFM(oogl_streams[i], TS[i], T1, P, C, G, Gr)
#endif
  }
  /*---------------------------------------------------------------------------*
   *           PHASE I:  Thermalize the structure                              *
   *---------------------------------------------------------------------------*/

  /* 
   * Note about configurations:  6/1/98
   * T1 is the configuration we use to bootstrap the parallel tempering
   * routine. Once this is done, it is used as a working copy of the positions
   * to be used inside the biased monte carlo routine.
   *
   */


  push_verbose_message("parallel_tempering","PHASE I: thermalize the structures");

  for(i = 0; i < n_betas; i++) {
    unsigned j;

    for(j = 0; j < P->sw_therm; j++) {
      int jj;
      /* ugly workaround */
      P->radius = radii[i];
      P->temperature = betas[i];
   
      fractions[i] = biased_monte_carlo(P, C, G, RP, TS[i], T1, Gr, Current[i]);
#ifdef ENERGY
      fprintf(ene_streams[i],"%d ", i);
      fprint_Conf(Current[i], ene_streams[i]);
      for(jj = 0; jj < TS[i]->Nmol; jj++) 
	print_Molecule(TS[i]->M[jj], mol_streams[i]);
#endif
      }
    
    sprintf(line, "Therm %d: T = %f fraction %f radius %f", i,
	    P->temperature, fractions[i], P->radius);
    push_verbose_message("parallel_tempering", line);
    
#ifdef OOGL
    OOGLCONFM(oogl_streams[i], TS[i], T1, P, C, G, Gr)
#endif
      
  }  

  /*---------------------------------------------------------------------------*
   *                      PHASE II: Parallel Tempering                         *
   *---------------------------------------------------------------------------*/


  push_verbose_message("parallel_tempering","PHASE II: tempering the structures");
  sprintf(line, "will do %ld PT steps",   
               P->granularity * P->sw_anneal * n_betas);
  push_verbose_message("parallel_tempering", line);

  for(i = 0 ; i <  P->sw_anneal * n_betas; i++) {
    int jj;
    unsigned pick;

    /* here we decide which action to take: update or swap */
    if(Random() < update_ratio) { /* do an update */

      pick = pick_system(rel_probs, n_betas);
      P->radius = radii[pick];
      P->temperature = betas[pick];

      fractions[pick] = 
	biased_monte_carlo(P, C, G, RP, TS[pick], T1, Gr, Current[pick]);
#ifdef ENERGY
      fprintf(ene_streams[pick],"%d ", i);
      fprint_Conf(Current[pick], ene_streams[pick]);
/* MWD Change */
      for(jj = 0; jj < TS[pick]->Nmol; jj++) 
	print_Molecule(TS[pick]->M[jj], mol_streams[pick]);
#endif
    } else { /* attempt to swap: note we pick fewer cases */
      int jj;
      double new, old, delta;
      double ene1, ene2;

      pick = floor((n_betas - 1) * Random());
      ene1 = Current[pick]->tot_energy;
      ene2 = Current[pick + 1]->tot_energy;
      old = ene1 / betas[pick] + ene2 / betas[pick + 1];
      new = ene2 / betas[pick] + ene1 / betas[pick + 1];
      delta = new - old;

      if(Random() < exp(-delta)) { /* swap configurations */
	struct Positions * T_tmp = TS[pick];
	struct Conf * C_tmp = Current[pick];

	TS[pick] = TS[pick + 1];
	Current[pick] = Current[pick + 1];
	TS[pick + 1] = T_tmp;
	Current[pick + 1] = C_tmp;

      
#ifdef ENERGY
	fprintf(ene_streams[pick], "# step %d: swapped with %d\n%d ", 
		i, pick + 1, i);
	fprint_Conf(Current[pick], ene_streams[pick]);
        /* MWD Change */
	if (T1->Nmol) 
          fprintf(mol_streams[pick], "# step %d: swapped with %d\n%d ", 
		i, pick + 1, i);
	for(jj = 0; jj < TS[pick]->Nmol; jj++) 
	  print_Molecule(TS[pick]->M[jj], mol_streams[pick]);
	
	fprintf(ene_streams[pick + 1], 
		"# step %d: swapped with %d\n%d ", i, pick, i);
	fprint_Conf(Current[pick + 1], ene_streams[pick + 1]);
        if (T1->Nmol)
          fprintf(mol_streams[pick + 1], 
		"# step %d: swapped with %d\n%d ", i, pick, i);
	for(jj = 0; jj < TS[pick + 1]->Nmol; jj++) 
	  print_Molecule(TS[pick + 1]->M[jj], mol_streams[pick + 1]);
#endif
      }
    }
    if (pick == 0) {
      sprintf(message,
              "s %d temp %e radius %e fraction %e energies:"
              " Tot %e RP %e Geom %e",
              i, betas[pick], radii[pick], fractions[pick],
              Current[pick]->tot_energy, Current[pick]->RP_energy,
              Current[pick]->Pos_energy);
      push_verbose_message("parallel_tempering",message);
      push_verbose_message("parallel_tempering", "positions");
      if(verbose_stream())  {
        write_unique_Positions(TS[pick], verbose_stream());
        write_Molecule_atoms(TS[pick]->Nmol, TS[pick]->M, verbose_stream());
      }
    }
#ifdef OOGL
    OOGLCONFM(oogl_streams[pick], TS[pick], T1, P, C, G, Gr)
#endif
  }

#ifdef OOGL
  for(i = 0; i < n_betas; i++) {
    fclose (oogl_streams[i]);
  }
  Free(oogl_streams);
#endif
#ifdef ENERGY
  for(i = 0; i < n_betas; i++) {
    fclose (ene_streams[i]);
    if(T1->Nmol) fclose(mol_streams[i]);
  }
  Free(ene_streams);
#endif

  delete_Grid(Gr);

  for(i=0; i < n_betas; i++) {
    delete_Positions(TS[i]);
    delete_Conf(Current[i]);
  }

  Free(TS);
  Free(Current);

  Free(betas);
  Free(radii);
  Free(phi);
  Free(fractions);
  Free(rel_probs);

  return;

}


/* this routine takes an array of doubles and returns a random array index, 
   according to the values of the array, supposedly normalized to 1 */

unsigned pick_system(double * P, unsigned max_index)
{
  unsigned i = 0;
  double pick = Random();
  double cump = P[0];

  while(cump < pick && i < max_index) {
    i++;
    cump += P[i];
  }

  /* make sure we do not over run the array */

  return (i < max_index ? i : max_index - 1);

} 
