/*

  here we deal with atom positions.
  the unique atoms are stored in the first n_unique rows of
  the coords matrix.  The rest of the matrix is used to
  store the symmetric copies when needed.

  Question: how do we deal with special positions?



  (C) 1997 Marco Falcioni (falcioni@ucla.edu)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
  USA.

  http://www.fsf.org/copyleft/gpl.html



*/

/* includes */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <ctype.h>

/* from libgjm-1.3 */
#include "memory.h"
#include "error.h" 


#include "atoms.h"
#include "group.h"
#include "symmetry.h"
#include "cell.h"
#include "chain.h"
#include "link.h"
#include "misc.h"
#include "verbose.h"
#include "positions.h"
#include "molecule.h"

/*---------------------------------------------------------------------------*
 *                          allocation: standard stuff                       *
 *---------------------------------------------------------------------------*/

struct Positions * new_Positions(unsigned long n_unique, 
				 unsigned long n_symmetry,
				 unsigned long n_actual,
                                 unsigned long Nmol,
                                 struct Molecule ** M)
{
  unsigned long i;
  struct Positions * T;

  T = (struct Positions *) Malloc(sizeof(struct Positions));

  if(n_actual > n_unique*n_symmetry) {
    push_generic_error("new_Positions","n_symmetries actual");
    exit_clean();
  }

  T->n_unique = n_unique;
  T->n_symmetry = n_symmetry;
  T->n_actual = n_actual;


  T->mod = 0;
  T->merged = 0;
  T->grid = 0;

  T->unique = new_Chain();
  T->distinct = new_Chain();

  T->sym_related = 
    (struct Chain **) Malloc(n_unique * sizeof(struct Chain *));

  for(i=0; i< n_unique; i++){
    unsigned long j;
    struct Link * L_unique;    
    struct Link * L_sym;
    struct Atom * A;

    T->sym_related[i] = new_Chain();

    A = new_Atom(); /* unique Atom */
    

    set_label_Atom(A,"None"); /* default action */
    A->root = (int) i;

    L_unique = new_Link(A);
    L_sym = new_Link(A);

    add_Link(L_unique, T->unique);
    add_Link(L_sym, T->sym_related[i]); /* add unique positions here */
    for(j = 1; j < n_symmetry; j++){
      A = new_Atom();
      set_label_Atom(A,"None");
      A->root = (int) i;
      L_sym = new_Link(A);
      add_Link(L_sym,T->sym_related[i]);
    }
  }

/* MWD Change */
  T->M = NULL;
  if(Nmol) T->M = (struct Molecule **) malloc(Nmol *
			  sizeof(struct Molecule *));
  for (i = 0; i < Nmol; i++){
    T->M[i] = new_Molecule(M[i]->N);
  }
  T->Nmol = Nmol;
  T->Nmol_atoms= 0;

  return T;
}

/*---------------------------------------------------------------------------*
 *          construct a copy of a set of positions                           *
 *---------------------------------------------------------------------------*/

void copy_Positions(const struct Positions * T_o, struct Positions * T_n) 
{
  unsigned i;
  struct Link * A;
  struct Link * B;

#ifndef NOCHECKS
  if(T_o == 0 || T_n == 0) {
    push_generic_error("copy_Positions", "(nil) arg passed");
    exit_clean();
  }

  if(T_o->n_unique != T_n->n_unique || T_o->n_symmetry != T_n->n_symmetry) {
    push_generic_error("copy_Positions", "incongruent arguments");
    exit_clean();
  }
#endif

  A = T_o->unique->head;
  B = T_n->unique->head;

  for(i = 0; i < T_o->n_unique; i++) {
    double pos[3];

    pos[0] = A->atom->ra;
    pos[1] = A->atom->rb;
    pos[2] = A->atom->rc;

    move_Atom(B->atom, pos);

    A = A->next;
    B = B->next;
  }


/* MWD Change */
  for(i = 0; i < T_o->Nmol; i++) {
    copy_Molecule(T_o->M[i], T_n->M[i]);
  }

  T_n->Nmol_atoms= T_o->Nmol_atoms;
  T_n->Nmol = T_o->Nmol;

  T_n->mod = 0;
  T_n->test = 0;
  T_n->merged = 0;
  T_n->grid = 0;

}


/*---------------------------------------------------------------------------*
 *          clone a copy of a set of positions                           *
 *---------------------------------------------------------------------------*/

void clone_Positions(const struct Positions * T_o, struct Positions * T_n) 
{
  unsigned i;
  struct Link * A;
  struct Link * B;

#ifndef NOCHECKS
  if(T_o == 0 || T_n == 0) {
    push_generic_error("copy_Positions", "(nil) arg passed");
    exit_clean();
  }

  if(T_o->n_unique != T_n->n_unique || T_o->n_symmetry != T_n->n_symmetry) {
    push_generic_error("copy_Positions", "incongruent arguments");
    exit_clean();
  }
#endif

  A = T_o->unique->head;
  B = T_n->unique->head;

  for(i = 0; i < T_o->n_unique; i++) {
    struct Link * bb;
    double pos[3];

    pos[0] = A->atom->ra;
    pos[1] = A->atom->rb;
    pos[2] = A->atom->rc;

    move_Atom(B->atom, pos);
    clone_Atom(B->atom, A->atom);

    /* copy also the symmetry related properties */
    for(bb = T_n->sym_related[i]->head->next; bb != 0; bb = bb->next) {
      clone_Atom(bb->atom, A->atom);
    }

    A = A->next;
    B = B->next;

  }


/* MWD Change
  if(T_o->Nmol) T_n->M = (struct Molecule **) malloc(T_o->Nmol *
			  sizeof(struct Molecule *));
*/


  for(i = 0; i < T_o->Nmol; i++) {
/* MWD Change
    T_n->M[i] = new_Molecule(T_o->M[i]->N); 
*/
    copy_Molecule(T_o->M[i], T_n->M[i]);
  }

  T_n->Nmol_atoms= T_o->Nmol_atoms;
  T_n->Nmol = T_o->Nmol;

  T_n->mod = 0;
  T_n->test = 0;
  T_n->merged = 0;
  T_n->grid = 0;

}

/*---------------------------------------------------------------------------*
 *               free a set of positions                                     *
 *---------------------------------------------------------------------------*/

void delete_Positions (struct Positions * T)
{
  int i;
  if(T == 0) return;
  clean_Chain(T->unique);
  delete_Chain(T->unique); 
  clean_Chain(T->distinct);
  delete_Chain(T->distinct);
  for(i = 0; i < T->n_unique; i++){
    clean_Chain(T->sym_related[i]);
    delete_Chain(T->sym_related[i]);
  }
  Free(T->sym_related);
  for(i = 0; i < T->Nmol; i++) 
    delete_Molecule(T->M[i]);
/* MWD Change */
  Free(T->M);
  Free(T);
  return;
}

/*---------------------------------------------------------------------------*
 *     init_Positions from stream, listing parameters and the                *
 *                   coordination of the unique T-atoms                      *
 *---------------------------------------------------------------------------*/

struct Positions * init_Positions (FILE * stream) 
{
  char line[1024];
  char * p;
  unsigned long c, size, a[3];
  long * bonds;

  struct Positions * T;
  struct Link * aa;
  
  if(stream == 0) {
    push_generic_error("init_Positions","(nil) arg passed");
    exit_clean();
  }

  push_verbose_message("init_Positions","initializing positions");

  /* get line strips white lines and cleans */

  p = get_line(line,sizeof(line),stream);
  
  if(p == 0 || 
     sscanf(clean_line(p), "%ld %ld %ld", &a[0], &a[1], &a[2]) != 3) {
    push_generic_error("init_Positions","problem reading data from stream");
    exit_clean();
  }

  p = get_line(line,sizeof(line),stream);

  size = count_cols(p); /* no need to clean the line */

  if(size == 0 || size != a[0]) {
    push_generic_error("init_Positions",
		       "bond coordination line does not match # of T-atoms");
    exit_clean();
  }

  bonds = read_longs(size,line);
  
  T = new_Positions(a[0],a[1],a[2], 0, NULL);

  /* assign the unique positions */
  c = 0;
  for(aa = T->unique->head; aa != 0; aa = aa->next) {
    unsigned i;
    double a[3];
    for(i = 0; i < 3; i++) a[i] = Random();
    move_Atom(aa->atom, a);
    aa->atom->n_bonds = (unsigned) bonds[c];
    c++;
  }

  Free(bonds);

  T->test = 0; /* the positions are not symmetrized, since we need
		  a symmetry group */
  T->mod = 0;
  T->merged = 0;
  T->grid = 0;

  return T;
}

/*---------------------------------------------------------------------------*
 *               randomize positions                                         *
 *---------------------------------------------------------------------------*/

void rand_Positions(struct Positions * T)
{
  struct Link * aa;

  if(T == 0) {
    push_generic_error("rand_Positions",
		       "(nil) arg passed or Positions not symmetric");
    exit_clean();
  }

  /* so easy using lists! */
  for(aa = T->unique->head; aa != 0; aa = aa->next) {
    double pos[3];

    pos[0] = Random();
    pos[1] = Random();
    pos[2] = Random();

    move_Atom(aa->atom, pos);
  }

  return;
}

/*---------------------------------------------------------------------------*
 *                put all positions in the unit cell                         *
 *---------------------------------------------------------------------------*/

void mod_Positions(struct Positions * T)
{
  unsigned long i;
  struct Link * aa;
  if(T->mod == 1) return;
  T->mod = 1;

  if(T == 0) {
    push_generic_error("mod_Positions",
		       "(nil) arg passed or Positions not symmetric");
    exit_clean();
  }

  /* so easy using lists! */
  for(i=0; i < T->n_unique; i ++) {
    for(aa = T->sym_related[i]->head; aa != 0; aa = aa->next) {
      mod_Atom(aa->atom);
    }
  }
  return;
}

/*---------------------------------------------------------------------------*
 *       print positions for debugging purposes                              *
 *---------------------------------------------------------------------------*/

void print_Positions (const struct Positions * T, FILE * stream)
{
  unsigned long i;
  if(T == 0 || stream == 0){
    push_generic_error("print_Positions",
		       "(nil) arg passed");
    exit_clean();
  }

  fprintf(stream,"Positions at %p {\n",T);
  fprintf(stream,"Unique Atoms at %p\n",T->unique);
  for(i = 0; i < T->n_unique; i++) {
    fprintf(stream,"Symmetry Related Atoms at %p\n",T->sym_related[i]);
  }
  fprintf(stream,"Distinct Atoms at %p\n",T->distinct);
  fprintf(stream,"n_unique %ld n_symmetry %ld n_actual %ld density %d\n",
	  T->n_unique, T->n_symmetry, T->n_actual, T->density);

  fprintf(stream,"} end Positions\n");
  return;
}

/*---------------------------------------------------------------------------*
 *        dump the abc positions of all the atoms in the unit cell           *
 *---------------------------------------------------------------------------*/

void write_Positions(const struct Positions * T, FILE * stream)
{
  unsigned long i;
  struct Link * aa;
  for(i=0; i < T->n_unique; i ++) {
    for(aa = T->sym_related[i]->head; aa != 0; aa = aa->next) {
      fprintf(stream, "%f %f %f\n",
	      aa->atom->ra, aa->atom->rb, aa->atom->rc);
    }
  }
  return;
}

/*---------------------------------------------------------------------------*
 *       dump the abc coordinates of the unique atoms (asymmetric subcell)   *
 *---------------------------------------------------------------------------*/

void write_unique_Positions(const struct Positions * T, FILE * stream)
{
  struct Link * aa;

  for(aa = T->unique->head; aa != 0; aa = aa->next) {
    fprintf(stream, "%4s %f %f %f %7.4f %7.4f\n",
	    aa->atom->label, aa->atom->ra, aa->atom->rb, aa->atom->rc,
	    aa->atom->DW_fac, aa->atom->occupancy);
  }
  return;  
}

/*---------------------------------------------------------------------------*
 *       write the positions in XTL format, fit for Cerius2  consumption     *
 *---------------------------------------------------------------------------*/

void write_unique_Positions_XTL(const struct Positions * T, FILE * stream)
{
  struct Link * aa;
  int i = 0;

  if(T == 0 || stream == 0){
    push_generic_error("write_unique_Positions_XTL", "(nil) arg passed");
    exit_clean();
  }

  fprintf(stream,"ATOMS\n");
  fprintf(stream,
	  "NAME       X          Y          Z     CHARGE   TEMP    OCCUP   SCAT\n");

  for(aa = T->unique->head; aa != 0; aa = aa->next) {
    fprintf(stream, "%2s%d   %8.5f    %8.5f    %8.5f  0.0000 %7.4f %7.4f   %2s0+\n",
	    aa->atom->label,i, aa->atom->ra, aa->atom->rb, aa->atom->rc, aa->atom->DW_fac, 
	    aa->atom->occupancy, aa->atom->label);
    ++i;
  }

  return;  
}

/*---------------------------------------------------------------------------*
 *      write the positions to a character array. do we need this?           *
 *---------------------------------------------------------------------------*/

char * sprint_Positions(const struct Positions * T, char * Ch)
{
  struct Link * aa;
  if(T == 0 || Ch == 0){
    push_generic_error("sprint_Positions", "(nil) arg passed");
    exit_clean();
  }
  for(aa = T->unique->head; aa != 0; aa = aa->next) {
    sprintf(Ch, "%f %f %f\n",
	    aa->atom->ra, aa->atom->rb, aa->atom->rc);
  }
  return Ch;  
}

/*---------------------------------------------------------------------------*
 *         print only the unique positions for debugging purposes            *
 *---------------------------------------------------------------------------*/

void print_unique_Positions (const struct Positions * T, FILE * stream)
{
  if(T == 0 || T->test != 1 || stream == 0){
    push_generic_error("print_unique_Positions",
		       "(nil) arg passed or Positions not symmetric");
    exit_clean();
  }
  fprintf(stream,"Positions at %p {\n",T);
  fprintf(stream,"Unique Atoms at %p {\n",T->unique);
  print_Chain(T->unique, stream);
  fprintf(stream,"} end Positions\n");
  return;
}

/*---------------------------------------------------------------------------*
 *     reads all unique positions from stream.                               *
 *     comments are allowed after the data, like this # comment              *
 *      but no blank or comment lines are allowed in the data                *
 *---------------------------------------------------------------------------*/

void read_Positions(struct Positions * T, FILE * stream)
{
  char line[1024];
  char * p;
  struct Link * aa;
  int i = 0;

  if(T == 0 || stream == 0) {
    push_generic_error("read_Positions", "(nil) arg passed");
    exit_clean();
  }

  for(aa = T->unique->head; aa != 0; aa = aa->next) {
    struct Link * bb;
    double x[3], DW, occ;
    int c;
    char label[10];
    p = get_line(line, sizeof(line), stream);
    
    c = sscanf(p, "%s %lf %lf %lf %lf %lf",
	       label, &x[0],&x[1],&x[2], &DW, &occ);

    if(c < 3) {
      push_generic_error("read_Positions",
			 "invalid T-atom position read from stream" );
      exit_clean();
    }

    if(c == 3) { /* old fashioned data */
      x[2] = x[1];
      x[1] = x[0];
      x[0] = strtod(label, 0);
      strcpy(label, "SI"); 
    }

    if(c >= 5) aa->atom->DW_fac = DW;
    if(c >= 6) aa->atom->occupancy = occ;
    
    move_Atom(aa->atom,x);
    set_label_Atom(aa->atom,label); /* default action */

    for(bb = T->sym_related[i]->head->next; bb != 0; bb = bb->next) {
      set_label_Atom(bb->atom, aa->atom->label); /* default action */
      bb->atom->DW_fac = aa->atom->DW_fac;
      bb->atom->occupancy = aa->atom->occupancy;
      bb->atom->n_bonds = aa->atom->n_bonds; /* crucial !!! */
      bb->atom->root = i; /* this is crucial as well */
    }

    ++i;
  }

  T->test = 0; /* not symmetric yet */

  p = get_line(line, sizeof(line), stream);
  if(p == 0) {
    push_generic_error("read_positions", "if you want to not see this"
		       " message\n put END at the end of the input file");
    return;                      /* no molecules in input */
  }

  if(strncmp("END", p, 3) == 0) return;

  sscanf(p, "%d", &(T->Nmol));
  
  T->M = (struct Molecule **) malloc(T->Nmol * sizeof(struct Molecule *));

  for(i = 0; i < T->Nmol; i++) {
    T->M[i] = read_Molecule(stream);
    T->Nmol_atoms += T->M[i]->N;
  }

  return;
}


/*---------------------------------------------------------------------------*
 *       read positions from a XTL data file                                 *
 *---------------------------------------------------------------------------*/

void read_Positions_XTL(struct Positions * T, FILE * stream)
{
  char line[1024];
  char * p;
  struct Link * aa;

  if(T == 0 || stream == 0) {
    push_generic_error("read_Positions_XTL", "(nil) arg passed");
    exit_clean();
  }

  /* read first line */
  p = get_line(line, sizeof(line), stream);
  if(strncmp(p,"ATOMS", (size_t) 5)) {
    push_generic_error("read_Positions_XTL", "error reading positions from stream 1");
    exit_clean();
  }

  p = get_line(line, sizeof(line), stream);
  if(strncmp(p,"NAME", (size_t) 4)) {
    push_generic_error("read_Positions_XTL", "error reading positions from stream 2");
    exit_clean();
  }

  for(aa = T->unique->head; aa != 0; aa = aa->next) {
    double x[3], m,n,o;
    char label[10], scra[10];
    int i = 0;

    p = get_line(line, sizeof(line), stream);
    if(p == 0 || 
       sscanf(p, "%s %lf %lf %lf %lf %lf %lf %s", 
	      label, &x[0],&x[1],&x[2], &m, &n, &o, scra) != 8 ){
      push_generic_error("read_Positions_XTL",
			 "invalid T-atom position read from stream" );
      exit_clean();
    }
    
    if(n > 0.0) aa->atom->DW_fac = n;

    aa->atom->occupancy = o;

    move_Atom(aa->atom,x);

    /* remove number from the atom label and uppercase letters */

    while(isalpha(label[i])) {
      int a = label[i];
      if(a < 123 && a > 96) /* a is lower case */ 
	label[i] -= 32;
      ++i;
    }

    label[i] = '\0';

    set_label_Atom(aa->atom,label); /* default action */
  }
  T->test = 0; /* not symmetric yet */
  return;
}


