/* gjstrings.c
 *
 * Created: [GJ] Sat Jul 15 19:12:04 EDT 1995
 *
 * [GJ] Fri Apr 26 04:22:33 EDT 1996
 * Added string array functions. Cleaned up interfaces.
 * Implemented general non-destruction policy.
 * 
 * [GJ] Thu Mar 20 01:00:11 EST 1997
 * Completely re-worked interface.
 *
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#ifdef _HAVE_STRING_H
#include <string.h>
#endif
#include <strings.h>
#include "io.h"
#include "resources.h"
#include "stringlib.h"


char * new_string(unsigned long size)
{
  char * s = (char *)malloc((size_t) size * sizeof(char));
  if(s==(char *)0){
    fprintf(stderr,"%s out of memory for a string\n", get_app_name());
  }
  return s;
}


void free_string(char * s)
{
  if(s != 0) free(s);
}


char * copy_string(const char * str)
{
  if(str==(char *)0){
    return (char *)0;
  }
  else {
    char * s = new_string((unsigned long)strlen(str)+1);
    if(strcpy(s, str)==(char *)0){
      fprintf(stderr,"%s: trouble copying a string\n", get_app_name());
      fprintf(stderr,"%s: probably a BUG\n", get_app_name());
    }
    return s;
  }
}


int in_string(int ch, const char * str)
{
  const char * sc;
  for(sc=str; *sc != '\0'; sc++)
    if((int) *sc == ch) return 1;

  return 0;
}


char * append_char_string(char * str, char ch)
{
  if(str == (char *)0){
    char *result = new_string(2);
    result[0] = ch;
    result[1] = '\0';
    return result;
  }
  else {
    int len = strlen(str) + 2;
    char *app = new_string(2);
    char *result = new_string(len);

    app[0] = ch;
    app[1] = '\0';

    strcpy(result, str);
    strcat(result, app);

    free_string(str);
    free_string(app);

    return result;
  }
}


void charsub_string(char * str, const char * badlist, char ch)
{
  if(str != (char *)0){
    int len = strlen(str);
    int i;

    for(i=0; i<=len; i++){
      if(in_string(str[i],badlist)) {
	str[i] = ch;
      }
    }
  }
}


void mangle_string(char *s)
{
  if(s == (char *)0) return;

  while(*s != '\0'){
    if(*s == ' ')  *s = '\a';
    if(*s == '\n') *s = '\b';
    s++;
  }
}


void unmangle_string(char *s)
{
  if(s == (char *)0) return;

  while(*s != '\0'){
    if(*s == '\a' || *s == '\b')  *s = ' ';
    s++;
  }
}


char * concat_strings(char * s1, const char * s2)
{
  char *s;
  int len;

  if(s1 == (char *)0 && s2 == (char *)0) 
    return (char *)0;
  
  if(s2 == (char *)0)
    return s1;

  if(s1 == (char *)0)
    return copy_string(s2);

  len = strlen(s1) + strlen(s2) + 1;
  s = new_string(len * sizeof(char));
  if(s == (char *)0){
    fprintf(stderr,"%s: no memory for concat_strings()\n", get_app_name());
    return 0;
  }
  
  strcpy(s, s1);
  return strcat(s, s2);
}


char * append_string(char *str1, const char *str2)
{
  char * newstring = concat_strings(str1, str2);
  free_string(str1);
  return newstring;
}


struct string_list * new_string_list(void)
{
  struct string_list *sl = (struct string_list *)
    malloc(sizeof(struct string_list));

  if(sl == (struct string_list *)0){
    fprintf(stderr,"%s: could not allocate a string_list\n",
	    get_app_name());
  }
  
  sl->size = 0;
  sl->str = (char **)0;

  return sl;
}


static struct string_list * create_string_list(int size)
{
  int i;
  struct string_list * sl = new_string_list();

  sl->str = (char **) malloc((size_t)(size * sizeof(char *)));
  if(sl->str == (char **)0){
    fprintf(stderr,"%s: could not allocate a string_list of size %d\n",
	    get_app_name(), size);
    exit(1);
  }
  
  sl->size = size;
  for(i=0; i<size; i++){
    sl->str[i] = (char *)0;
  }

  return sl;
}


struct string_list *create_string_list_from_array(const char **a, int size)
{
  int i;
  struct string_list *sl = create_string_list(size);

  for(i=0; i<size; i++){
    sl->str[i] = copy_string(a[i]);
  }

  return sl;
}


void free_string_list(struct string_list * sl)
{
  int i;
  if(sl != (struct string_list *)0){
    for(i=0; i<sl->size; i++){
      /* if(sl->str[i] != (char *)0) */ free_string(sl->str[i]);
    }
    free(sl);
  }
}


struct string_list * copy_string_list(const struct string_list * sl)
{
  int i;
  struct string_list * newsl = create_string_list(sl->size);
  for(i=0; i<sl->size; i++){
    newsl->str[i] = copy_string(sl->str[i]);
  }

  return newsl;
}


struct string_list * append_string_to_list(struct string_list *sl,
					   const char *str)
{
  int i;
  struct string_list *newsl;

  if(str == (char *)0) return sl;

  if(sl == (struct string_list *)0){
    newsl = create_string_list(1);
    newsl->str[0] = copy_string(str);
    newsl->size = 1;
  }
  else {
    newsl = create_string_list(sl->size + 1);
    for(i=0; i<sl->size; i++){
      newsl->str[i] = sl->str[i];
    }
    newsl->str[sl->size] = copy_string(str);

    /* The contents of sl should not be destroyed since they
     * were simply shuffled into newsl. However sl itself
     * can disappear.
     */
    free(sl);
  }

  return newsl;
}


struct string_list * concat_string_lists(const struct string_list *sl1,
					 const struct string_list *sl2)
{
  int i;
  struct string_list *newsl = create_string_list(sl1->size + sl2->size);
  newsl->size = sl1->size + sl2->size;

  for(i=0; i<sl1->size; i++){
    newsl->str[i] = copy_string(sl1->str[i]);
  }
  for(i=0; i<sl2->size; i++){
    newsl->str[i+sl1->size] = copy_string(sl2->str[i]);
  }

  return newsl;
}


struct string_list * split_string(const char *str, const char *separators)
{
  int i;
  char *field = (char *)0;
  struct string_list *sl = new_string_list();

  if(str == (char *)0) return sl;

  for(i=0; str[i] != '\0'; i++){
    if(in_string(str[i],separators)){
      sl = append_string_to_list(sl, field);
      free_string(field);
      field = (char *)0;
    }
    else {
      field = append_char_string(field, str[i]);
    }
  }

  /* Add any left over that we were working on. */
  if(field != (char *)0){
    sl = append_string_to_list(sl, field);
    free_string(field);
  }

  return sl;
}


int in_string_list(const struct string_list *sl, const char *str)
{
  int i;

  if(sl == (struct string_list *)0) return 0;

  for(i=0; i<sl->size; i++){
    if(sl->str[i] != (char *)0){
      if(strcmp(sl->str[i], str) == 0) return 1;
    }
  }

  return 0;
}


double double_from_string(const char *str)
{
  double result = 0.;
  if(sscanf(str,"%lf",&result) != 1){
    fprintf(stderr,"%s: could not scan a double from string=%s\n",
	    get_app_name(), str);
  }
  return result;
}


int int_from_string(const char *str)
{
  int result = 0;
  if(sscanf(str,"%d",&result) != 1){
    fprintf(stderr,"%s: could not scan an int from string=%s\n",
	    get_app_name(), str);
  }
  return result;
}
